//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	AuxHKaccess.cc	--- implementation of Auxiliary/HK access classes
//
//	DESCRIPTION:
//		Implementation part of classes in AuxHKaccess.h
//
//	CHANGE HISTORY:
//		UL, 2001-06-15: 1.0.0: - first version
//		UL, 2001-06-20: 1.1.0: - major overhaul; extensive tests for EPN
//								 AUX anomalies
//		UL, 2001-07-24: 1.2.0: - check for unusable frames
//		UL, 2002-02-04: 1.3.0: - computation of frameTime (_frameTime)
//								 was failing in case of SmallWindowMode
//								 under some circumstances
//		UL, 2002-02-05: 1.3.1: - fixed minor compilation problem under
//								 optimization
//		UL, 2002-02-14: 1.4.0: - EpnAuxDataServer: in frame time computation,
//								 ignore first 0.1% of valid frames to
//								 exclude pathological cases
//		UL, 2002-02-21: 1.5.0: - optimized memory usage in EpnAuxDataServer
//								 constructor
//		UL, 2002-03-14: 1.6.0: - raised 0.1% (see 1.4.0 change) to 0.5% to
//								 rectify processing problem with ODF
//								 0375_0134521401
//		UL, 2002-03-08: 1.7.0: - EpnAuxDataServer::EpnAuxDataServer: fixed
//								 minor problem when second-last frame has
//								 FTCOARSE=32767
//		UL, 2002-04-02: 1.8.0: - fixed bug related to treatment of block
//								 of AUX data rows with wrong QUADRANT/CCDIDs
//								 [_accumReset vector element at block end
//								  was incorrect]
//
//	AUTHOR:
//		U. Lammers, Mon Aug  3 16:10:16 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/oal/src/AuxHKaccess.cc,v 1.16 2002/04/02 15:32:58 ulammers Exp $
//

#include <iostream>
#include <functional>
#include <algorithm>

#include "AuxHKaccess.h"

#include "Dal.h"
#include "Tracer.h"
#include "TimeTags.h"
#include "Xmm.h"
#include "Odf.h"
#include "OalServer.h"
#include "CalOalUtility.h"
#include "OalDalUtils.h"

using namespace CalOalUtility;
using namespace std;

HKParamFinder::HKParamFinder(const OalServer *server)
	: _server(server), _table(0)
{
	_instr = _server->instrument();
	const OdfSet *hkSet;
	if (_instr == Xmm::EMOS1 || _instr == Xmm::EMOS2)
		hkSet = server->odf()->selectSet(Xmm::PERIODIC_HOUSEKEEPING);
	else
		//	for PN the reset time is in the main periodic HK
		hkSet = server->odf()
							->selectSet(Xmm::MAIN_PERIODIC_HOUSEKEEPING);

	assert(hkSet);

	errstr<<"Opening periodic HK set `"
		  <<hkSet->name()<<"' to look for needed parameters"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	_table = dataSetServer->open(hkSet->name(), DataSet::Read,
								 DataSetServer::Low)->table(0);
	_index = 0;
} // HKParamFinder()


void HKParamFinder::seek(const STime &t, double deltaT = 0.) const
{
	if (!_table)
		return;

	double tt = t - Xmm::missionReferenceTime + deltaT;
	errstr<<"looking for HK parameter at t="<<tt<<endl
		   <<Sas::message(Msg::Lib, Msg::Verbose);

	const TimeCorrelator *timeCorrelator = _server->timeCorrelator();

	ColumnData *timeCol = _table->column("TIME")->data();

	unsigned nRows = _table->rows(), n;
	for (n=0; n<nRows; ++n) {
		string time(stringColumnRow(timeCol, n));
		double sec = timeCorrelator->toMissionEllapsedTime(time);
		errstr<<time<<" =^ "<<sec<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);
		if (sec >= tt)
			//	found row
			break;
	}
	delete timeCol;

	if (n >= nRows)
		errstr<<"Could not find time t="<<Xmm::missionReferenceTime+tt
			  <<" in periodic HK table `"
			  <<_table->name()<<"'"<<Sas::error("TimeMismatch");

	_index = n;
} // HKParamFinder::seek()


HKParamFinder::~HKParamFinder()
{
	if (_table)
		dataSetServer->close(_table->parent());
} // HKParamFinder::~HKParamFinder()


MainPeriodicHKParamFinder::MainPeriodicHKParamFinder(const OalServer *server)
		: HKParamFinder(server)
{
	const OdfSet *hkSet;
	if (_instr == Xmm::EMOS1 || _instr == Xmm::EMOS2)
		hkSet = server->odf()->selectSet(Xmm::PERIODIC_HOUSEKEEPING);
	else
		//	for PN the reset time is in the main periodic HK
		hkSet = server->odf()->selectSet(Xmm::MAIN_PERIODIC_HOUSEKEEPING);

	if (!hkSet)
		errstr<<"There is no main periodic HK file for this exposure"
			  <<Sas::error("MissingHKdata");

	errstr<<"Opening periodic HK set `"
		  <<hkSet->name()<<"' to look for required parameters"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	_table = dataSetServer->open(hkSet->name(), DataSet::Read,
								 DataSetServer::Low)->table(0);
	_index = 0;
} // MainHKParamFinder()


//
//	PURPOSE:
//		Get the values of the EPIC (MOS or PN) clock reset times from the
//		corresponding HK file.
//
//		The EPIC clock is reset once at the beginning of every
//		exposure and whenever the FTCORASE value wraps-around. There
//		can also be commanded resets during the exposure.
//		The times of these event are available as a COBT stamp
//		(Coarse+Fine Time) in the periodic HK file belonging to this
//		exposure. The function reads the value of reset time corresponding
//		to 30 sec after the start of the exposure. This is necessary
//		because sometimes the reset pulse becomes a couple of seconds
//		after the exposure start.
//
//	RETURNS:
//		EPIC clock reset times of current exposure as list of double COBT
//		values
//
vector<double> MainPeriodicHKParamFinder::resetTimes() const
{
	string ftCoarseName,
		   ftFineName;
	switch (_instr) {
		case Xmm::EMOS1: ftCoarseName	= "E1573";
						 ftFineName		= "EU573";
						 break;
		case Xmm::EMOS2: ftCoarseName	= "K1573";
						 ftFineName		= "KU573";
						 break;
		case Xmm::EPN:	 ftCoarseName	= "F1030";
						 ftFineName		= "FU036";
						 break;
		default:		 break;
	}

	ColumnData *cdc = _table->column(ftCoarseName)->data(),
			   *cdf = _table->column(ftFineName)->data();
	int32 *coarseTime = cdc->int32Data();
	uint16 *fineTime = cdf->uint16Data();

	STime tstart(_server->odf()->currentExposure()->actualStartTime()),
		  tend(_server->odf()->currentExposure()->actualEndTime());
	const TimeCorrelator *timeCorrelator = _server->timeCorrelator();

	unsigned N = _table->rows(), n=_index;
	vector<double> resetTimes;
	double		   rt;

	while (n<N) {
		int32 ct = coarseTime[n];
		rt = DBL(OalDalUtils::unsigned32(ct)) +
				ObtFineTimeResolution *
					DBL(OalDalUtils::unsigned32(fineTime[n]));

		//
		//	check that the reset time is not later than the exposure end
		//	time; if this IS the case there was probably a CDMU crash
		//	during the exposure and the reset time is wrong, i.e., shows
		//	the OBT value before the crash; if we continue with this
		//	wrong (too large) value all MET tags calculated from the TCS
		//	shall be wrong as well. Examples are all ODFs from revolutions
		//	103+104 (also F. Gianinni's email on CDMU crashes dated
		//	2001-03-29)
		//
		//	In this case we carry on with the OBT value corresponding
		//	to the actual exposure start - Note: This value can be
		//	off by a couple of seconds, i.e., all METs will be afflicted
		//	with the same absolute error.
		//
		STime treset(Xmm::missionReferenceTime +
					 timeCorrelator->toMissionEllapsedTime(rt));
		timeCorrelator->resetObtMetRelation();
		double deltaT = tend-treset;
		if (deltaT<0.) {
			errstr<<"The "<<Xmm::toString(_instr)<<" clock reset time read"
					" from HK parameters "<<ftCoarseName<<'/'<<ftFineName
				  <<" lies "<<-deltaT<<"s after the end of the current "
					"exposure at "<<tend<<". The most likely cause of this"
					" is a CDMU crash which occured during the exposure. "
					"Will set the reset time to the exposure start at "
				  <<tstart<<". Please note: This value can be inaccurate "
					"by several seconds and thus, all event time tags "
					"shall be afflicted with the same absolute error."
				  <<Sas::warning("WrongClockResetTime");
			rt = timeCorrelator->toOnBoardTime(tstart -
												Xmm::missionReferenceTime);
		}

		resetTimes.push_back(rt);
		errstr<<"EPIC clock reset time: "<<rt<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		//	search next reset time
		while (n<N && coarseTime[n] == ct)
			++n;
	}

	delete cdc;
	delete cdf;

	return resetTimes;
} // MainPeriodicHKParamFinder::resetTimes()


unsigned MainPeriodicHKParamFinder::wrapValue() const
{
	if (_instr != Xmm::EPN)
		return 32768;

	if (!_table->hasColumn("F1052")) {
		errstr<<"Expected Periodic HK parameter F1052 missing from `"
			  <<_table->name()<<"' - will continue with nominal value "
				"32400 as coarse time wrap value for EPN"
			  <<Sas::warning("NoSuchPeriodicHKparam");
		return 32400;
	}

	ColumnData *cdf1052 = _table->column("F1052")->data();
	unsigned val = cdf1052->uint16Data()[_index];
	delete cdf1052;

	return val;
} // MainPeriodicHKParamFinder::wrapValue()



AdditionalPeriodicHKParamFinder::AdditionalPeriodicHKParamFinder(
		const OalServer *server) : HKParamFinder(server)
{
	_table = 0;
	_index = 0;
	if (server->odf()->hasSet(Xmm::ADD_PERIODIC_HOUSEKEEPING)) {
		const OdfSet *hkSet = server->odf()->selectSet(
											Xmm::ADD_PERIODIC_HOUSEKEEPING);

		errstr<<"Opening periodic HK set `"
			  <<hkSet->name()<<"' to look for required parameters"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		_table = dataSetServer->open(hkSet->name(), DataSet::Read,
											 DataSetServer::Low)->table(0);
	} else
		errstr<<"Odf `"<<server->odf()->directoryName()<<"' contains no "
			  <<Xmm::toString(Xmm::ADD_PERIODIC_HOUSEKEEPING)<<" data - "
			    "will use default values for any needed parameters from this "
				"data set"<<Sas::warning("NoAdditionalPeriodicHKdata");
} // AdditionalPeriodicHKParamFinder::AdditionalPeriodicHKParamFinder()


unsigned short AdditionalPeriodicHKParamFinder::chopperValue(
											unsigned short quadrant) const
{
	unsigned short val = 0;

	if (_instr != Xmm::EPN)
		return 0;

	if (_table) {
		const char *colNames[] = { "F1534", "F1634", "F1734", "F1834" };
		assert(quadrant < NUM_OF(colNames));

		if (_table->hasColumn(colNames[quadrant])) {
			ColumnData *cdf = _table->column(colNames[quadrant])->data();
			val = cdf->int8Data()[_index];
			delete cdf;

			return val;
		}
	}

	errstr<<"Using default value 0 for missing additional periodic HK "
			"parameters F1534/F1634/F1734/F1834"
		  <<Sas::warning("DefaultValueForMissingParameter");

	return val;
} // AdditionalPeriodicHKParamFinder::chopperValue()


EpnAuxDataServer::EpnAuxDataServer(const OalServer *oals,
								   unsigned short quadrant,
								   unsigned short ccd,
								   unsigned short chopper,
								   unsigned wrapValue,
								   vector<double> resetT)
	: _quadrant(quadrant), _ccd(ccd), _chopper(chopper), _wrapValue(wrapValue),
	  _resetT(resetT)
{
	const OdfSet *odfSet = oals->odf()->selectSet(Xmm::AUXILIARY);

	if (!odfSet)
		errstr<<"Could not locate PN auxiliary file associated with "
				"current exposure of ODF in `"
			  <<oals->odf()->directoryName()<<"'"
			  <<Sas::error("NoAuxFile");

	DataSet *auxSet = dataSetServer->open(odfSet->name(), DataSet::Read,
									  DataSetServer::Low);
	Table *auxTable = auxSet->table(0);
	_tableName = auxTable->qualifiedName();
	ColumnData *cdframe = auxTable->column("FRAME")->data(),
			   *cdccdId = auxTable->column("CCDID")->data(),
			   *cdquadId = auxTable->column("QUADRANT")->data(),
			   *cdftcoarse = auxTable->column("FTCOARSE")->data(),
			   *cdftfine = auxTable->column("FTFINE")->data();
	int32	*frame = cdframe->int32Data();
	int8	*ccdId = cdccdId->int8Data(),
			*quadId = cdquadId->int8Data();
	int16	*ftcoarse = cdftcoarse->int16Data();
	uint16	*ftfine = cdftfine->uint16Data();

	unsigned N = auxTable->rows();
	_frameTime = 0.;
	double	 ftmult = DBL(chopper+1);
	unsigned NF = 0;
	for (unsigned n=0; n<N; ++n)
		if (ccdId[n] == ccd && quadId[n] == quadrant)
			++NF;

	++NF;	//	one more
	vector<double> fv;
	fv.reserve(NF);
	_frame.reserve(NF);
	_ftcoarse.reserve(NF);
	_ftfine.reserve(NF);
	_status.reserve(NF);
	_accumResets.reserve(NF);
	_deltaT.reserve(NF);

	unsigned k = 0;
	for (unsigned n=0; n<N; ++n) {
		if (ccdId[n] == ccd && quadId[n] == quadrant) {
			//	found frame
			_frame.push_back(frame[n]);
			_ftcoarse.push_back(ftcoarse[n]);
			_ftfine.push_back(ftfine[n]);
			_status.push_back(NORMAL);
			_accumResets.push_back(0);
			if (k) {
				double dt = getTime(k) - getTime(k-1);
				_deltaT.push_back(dt);
				if (dt > 0. &&	//	must be positive ...
					dt< 100.)	//	... and smaller than some maximum value
					fv.push_back(dt);
			}
			++k;
		}
	}

	delete cdframe;
	delete cdccdId;
	delete cdquadId;
	delete cdftcoarse;
	delete cdftfine;
	dataSetServer->close(auxSet);

	//	compute frame time as median of distribution in fv
	if (fv.empty())
		errstr<<"No usable frames in `"<<_tableName<<"' for quadrant/CCD="
			  <<quadrant<<'/'<<ccd<<Sas::error("NoUsableFrames");
	sort(fv.begin(), fv.end());
	N = fv.size();
	double median = fv[N/2],	//	most probable element
								//	smallest time > 0
		   minVal = 0.;

	const unsigned ignoreFirstFrames = UNT(.005*DBL(N));
	//	erase first 0.5% of all elements to get rid of 'pathological' values
	if (fv.size() > ignoreFirstFrames)
		fv.erase(fv.begin(), fv.begin()+ignoreFirstFrames);

	vector<double>::const_iterator i = find_if(fv.begin(), fv.end(),
										 bind2nd(greater<double>(), 0.));
	if (i != fv.end())
		minVal = *i;
	else
		errstr<<"No positive frame times found - something is very wrong"
			  <<Sas::error("NoPositiveFrameTimes");

	//	check if median is close to integer multipe of minVal
	double mult,
		   frac = modf(median/minVal, &mult);
	if (frac > .95)
		_frameTime = median/(mult+1.);
	else if (frac < .05)
		_frameTime = median/mult;
	else {
		//	something is wrong here: most probable time difference between
		//	consecutive frame does not appear to be an integer multiple of
		//	frame integration time; set frame time to minimum value
		//	can carry on
		errstr<<"Most probable time difference between consecutive frames "
				"does not appear to be an integer multiple of frame "
				"integration time - adjusting the latter to the minimum "
				"value of time differences"
			  <<Sas::warning("NotIntegerMultiple");
		_frameTime = minVal;
	}

	//	find first element 5% larger than nominal frame time and delete
	//	subsequent elements from vector
	vector<double>::iterator to = find_if(fv.begin(), fv.end(),
										  bind2nd(greater<double>(),
															1.05*_frameTime));
	fv.erase(to, fv.end());
	_frameTime = fv[fv.size()/2];
	fv.clear();

	_deltaT.push_back(_frameTime*ftmult);	//	fill last element

	//	add one element to the end
	N = _deltaT.size();
	_ftcoarse.push_back(0);
	_ftfine.push_back(0);
	_accumResets.push_back(0);
	setTime(N, getTime(N-1) + _deltaT[N-1]);
	_status.push_back(NORMAL);
	_frame.push_back(_frame.back() + 1);
	_deltaT.push_back(_frameTime*ftmult);

	//
	//	correct known anomalies
	//
	correctAnomalies();

	SAS_DEBUG(for (unsigned n=0; n<_deltaT.size(); ++n)
			  cout<<n<<' '<<_deltaT[n]/_frameTime<<endl;)
} // EpnAuxDataServer::EpnAuxDataServer()


int EpnAuxDataServer::locateFrame(int frameNr) const
{
	static int lastIndx = 0;

	vector<int>::const_iterator i = find(_frame.begin()+lastIndx,
													_frame.end(), frameNr);
	if (i == _frame.end()) {
		//	search again from the beginning
		i = find(_frame.begin(), _frame.end(), frameNr);
		if (i == _frame.end()) {
			lastIndx = 0;
			return -1;
		}
	}

	return lastIndx = i - _frame.begin();
} // EpnAuxDataServer::locateFrame()


int EpnAuxDataServer::nextNormalFrame(int start) const
{
	vector<unsigned char>::const_iterator i =
		find(_status.begin()+start, _status.end(), NORMAL);

	return i - _status.begin();
} // EpnAuxDataServer::nextNormalFrame()


double EpnAuxDataServer::getTime(unsigned i)
{
	return DBL(_ftcoarse[i]) + DBL(_ftfine[i]) * EpnFineTimeResolution;
} // EpnAuxDataServer::getTime()


double EpnAuxDataServer::getObt(unsigned i)
{
	return getTime(i) + _resetT[_accumResets[i]];
} // EpnAuxDataServer::getObt()


void EpnAuxDataServer::setTime(unsigned n, double t)
{
	double ft;
	_ftfine[n] = INT(fabs(modf(t, &ft))/EpnFineTimeResolution);
	_ftcoarse[n] = INT(ft);

	if (t<0.) {
		_ftfine[n] = INT(1./EpnFineTimeResolution-DBL(_ftfine[n])+.5);
		--_ftcoarse[n];
	}
		
} // EpnAuxDataServer::setTime()


//
//	maximum relative deviation from mean frame time allowed to
//	consider delta time between consecutive frames an integer multiple
//	of frame integration time
//
unsigned EpnAuxDataServer::deltaIsIntegerMultipleOfFrameTime(unsigned i) const
{
	//
	//	assign maximum allowed relative frame time deviation: 0.75%
	//
	const double maxRelativeFrameTimeDeviation = .0075;

	if (_deltaT[i]<=0.)
		return 0;

	double mult,
		   frac = modf(_deltaT[i]/_frameTime, &mult);

	if (frac > (1.-(mult+1.)*maxRelativeFrameTimeDeviation))
		return UNT(mult + 1.);

	if (frac > mult*maxRelativeFrameTimeDeviation)
		return 0;

	return UNT(mult);
} // EpnAuxDataServer::deltaIsIntegerMultipleOfFrameTime()


inline static int nearestInt(double x)
{
    return INT(x >= 0 ? floor(x + .5) : ceil(x - 0.5));
}

//
//	return integer i for which dt = n*FT+i, n smallest possible integer
//
int EpnAuxDataServer::integerOffsetFromMultipleFrameTime(double dt) const
{
	int di = nearestInt(dt/fabs(dt)),
		i = INT(dt);
	double dum;
	while (i) {
		double d = fabs(modf((dt - DBL(i))/_frameTime, &dum));
		if (d>1.-1.e-3 || d<1.e-3)
			break;
		i -= di;
	}

	return -i;
} // EpnAuxDataServer::integerOffsetFromMultipleFrameTime()


//
//	Try to correct the following anomalies in pn-AUX data
//
//		o spurious values in FTCOARSE
//			+ cause: buffer overflow in electronics
//			+ signature: FTCOARSE==32767
//			+ corrective action: approximate FTCOARSE by good time
//				of preceding [or following] frame plus [minus] nominal
//				frame integration time
//		o wrong FTCOARSE values for first frames
//			+ cause: reset pulse came too late
//			+ signature: first (at most three?) FTCOARSE values are
//				>0
//			+ corrective action: find first ok FTCOARSE value and
//				approximate all preceding values (+FTFINE) by this good
//				value minus integer multiple of nominal frame integration
//				time
//		o random negative or positive jumps in FTCOARSE
//			+ cause: unknown; bit-flips/SEU? (see email Fabio Gianinni
//				dated 12-06-2001)
//			+ signature: time between consecutive events is negative
//				or not an integer multiple of frame integration time;
//				all subsequent FTCOARSEs are wrong until the next reset
//			+ corrective action: 
//				adjust FTCOARSE value (and all subsequent ones until
//				next reset) to make time difference integer multiple of
//				nominal integration time
//
void EpnAuxDataServer::correctAnomalies()
{
	int N = _status.size() - 1;

	//	check the first ten frames for a negative time jump - if there
	//	is one the reset pulse arrived late - correct the coarse time
	//	in this case
	static const int framesToCheck = 10;	//	assume reset does not
											//	come later than 10 frames
											//	into the exposure

	int i;
	for (i=0; i<framesToCheck && i<N; ++i)
		if (_deltaT[i] < 0.)
			break;

	if (i+1>=N)
		errstr<<"Did not find any reliable FTCOARSE/FTFINE values "
				"in PN AUX data - this cannot happen - something is very "
				"wrong."<<Sas::fatal("RattyAuxData");

	if (i<framesToCheck) {
		for (int k=1, n=i; n>=0; --n, ++k) {
			_deltaT[n] = _deltaT[i+1];
			double t = getTime(i+1) - k*_deltaT[n];
			setTime(n, t);
		}
		_qualityLogger.lateResets.value = i+1;
	}

	//
	//	look for FTCOARSE==32767
	//
	int n = 0;
	while (n<N) {
		vector<int>::iterator
			from = find(_ftcoarse.begin()+n, _ftcoarse.end(), 32767),
			to = find_if(from, _ftcoarse.end(),
						 	  bind2nd(not_equal_to<int>(), 32767));
		if (to == _ftcoarse.end())
			break;

		int fromi = from - _ftcoarse.begin(),
			toi = to - _ftcoarse.begin() - 1;
		fill(_status.begin()+fromi, _status.begin()+toi+1, UNUSEABLE32767);
		_deltaT[fromi-1] = getTime(toi+1) - getTime(fromi-1);
											//	ignore this frame;
											//	set delta of
											//	preceding frame
		_qualityLogger.spurious32767s.value += fromi-toi+1;
		n = toi+1;
	}

	//
	//	look for positive/negative FTCOARSE jumps
	//
	double &min = _qualityLogger.minDelta,
		   &max = _qualityLogger.maxDelta;
	int tcorr = 0;
	unsigned nWraps = 0;
	for (int n=0; n<N; ++n) {
		_ftcoarse[n] += tcorr;
		_accumResets[n] = nWraps;
		if (_status[n] == NORMAL &&
			_deltaT[n] < 0. && -_deltaT[n] > DBL(_wrapValue/2)) {
			if (++nWraps >= _resetT.size()) {
					errstr<<"The frame coarse time has wrapped "<<nWraps+1
						  <<" time(s) in the pn auxiliary file `"
						  <<_tableName<<"' but there appear to have been "
							"only "<<_resetT.size()<<" clock reset(s) "
							"according to the HK data - will assume missing "
							"reset occured "<<_wrapValue<<" after last one"
						  <<Sas::warning("TooManyWrapArounds");
					_resetT.push_back(_resetT.back() + _wrapValue);
			}
			//	this is believed to be a true wrap around; from
			//	here onward the FTCOARSE values should be ok again
			_status[n] |= RESET;
			double deltaT = getTime(n+1) + _resetT[nWraps] -
						   (getTime(n) + _resetT[nWraps-1]);
			_qualityLogger.normalResets.incr(_ftcoarse[n], _ftfine[n],
											 deltaT);
			tcorr = 0;
			continue;
		}

		if (_status[n] == NORMAL) {
			int mult = deltaIsIntegerMultipleOfFrameTime(n);
			if (_deltaT[n] < 0. || !mult) {
				if (_deltaT[n+1] < 0. && _ftfine[n+1] >= MAX_FTFINE) {
					//
					//	FTCOARSE increment occured too early; this happens
					//	sometimes at FTFINE==MAX_FTFINE (=48828)
					//
					_qualityLogger.prematureIncrements.incr(_ftcoarse[n+1],
														    _ftfine[n+1]);
					_ftcoarse[n+1] -= 1;
					_deltaT[n] -= 1.;
					_deltaT[n+1] += 1.;
					continue;
				}

				int tc;
				if ((tc = integerOffsetFromMultipleFrameTime(_deltaT[n]))) {
					tcorr += tc;
					_deltaT[n] += tc;
					if (_deltaT[n] < 0)
						_qualityLogger.negativeJumps.incr(_ftcoarse[n],
														  _ftfine[n],
														  _deltaT[n] - tc);
					else
						_qualityLogger.positiveJumps.incr(_ftcoarse[n],
														  _ftfine[n],
														  _deltaT[n] - tc);
				} else {
					//
					//	check if there is a corresponding jumps in the
					//	the oposite direction later of the same magnitude
					//	some 25 frames later; if yes, this is a frame
					//	sequence that belongs to another quadrant/CCD.
					//
					vector<double>::iterator from = _deltaT.begin() + n+1,
												  to;
					//	time increment delta: try to find matching delta
					//	in opposite direction of value > .99*fabs(delta)
					double sign = _deltaT[n] < 0. ? 1. : -1.,
						   absMatchDelta = .99 * fabs(_deltaT[n]);
					to = find_if(from, _deltaT.end(),
							bind2nd(greater<double>(), sign*absMatchDelta));
					int fromi = from - _deltaT.begin(),
						frameDelta = to - _deltaT.begin() - 1 - fromi;
					if (to != _deltaT.end() && frameDelta>0 && frameDelta<25) {
						//	there was an associated jumps
						int toi = to - _deltaT.begin();
						fill(_status.begin()+fromi, _status.begin()+toi+1,
								  WRONG_QUAD_CCD);
						_deltaT[fromi-1] = getTime(toi+1) - getTime(fromi-1);
						_qualityLogger.wrongQuadCcds.value += toi-fromi+1;
						fill(_accumResets.begin()+fromi,
							 _accumResets.begin()+toi+1, nWraps);
						n = toi;
						continue; // execution will continue with element
								  // fromi+1
					} else
						_qualityLogger.unidentifiedDeltas.incr(_ftcoarse[n],
															   _ftfine[n],
															   _deltaT[n]);
				}
			} else if (mult > _chopper+1)
				_qualityLogger.missingFrames.value += mult - (_chopper+1);
			min = std::min(_deltaT[n], min);
			max = std::max(_deltaT[n], max);
		}
	}
} // EpnAuxDataServer::correctAnomalies()


EpnAuxDataServer::~EpnAuxDataServer()
{
	_ftcoarse.clear();
	_ftfine.clear();
	_accumResets.clear();
	_deltaT.clear();
	_status.clear();

	errstr<<"Discontinuities in auxiliary data from `"
		  <<_tableName<<"' for CCD/Quadrant: "<<_ccd<<'/'<<_quadrant
		  <<" ["<<EPNcam::toChipNr(_ccd, _quadrant)<<"] :"
		  <<Sas::message(Msg::Lib, Msg::Silent);
	errstr<<_qualityLogger<<endl
		  <<Sas::message(Msg::Lib, Msg::Silent);
} // EpnAuxDataServer::~EpnAuxDataServer()


ostream &operator<<(ostream &os, const EpnAuxDataServer::QualityLogger &ql)
{
	os<<endl
	  <<"    spurious 32767-frames: "<<ql.spurious32767s<<endl
	  <<"    late-reset frames    : "<<ql.lateResets<<endl
	  <<"    missing frames       : "<<ql.missingFrames<<endl
	  <<"    normal resets        : "<<ql.normalResets<<endl
	  <<"    premature increments : "<<ql.prematureIncrements<<endl
	  <<"    negative jumps       : "<<ql.negativeJumps<<endl
	  <<"    positive jumps       : "<<ql.positiveJumps<<endl
	  <<"    wrong quadrant/CCD   : "<<ql.wrongQuadCcds<<endl
	  <<"    unidentified deltas  : "<<ql.unidentifiedDeltas<<endl
	  <<endl
	  <<"    minimum delta        : "<<setprecision(4)<<ql.minDelta<<"s\n"
	  <<"    maximum delta        : "<<setprecision(4)<<ql.maxDelta<<'s';

	return os;
} // operator<<


ostream &operator<<(ostream &os, const EpnAuxDataServer::AnomalyCounter &ac)
{
	unsigned printMax = 100;

	os<<ac.value;
	if (!ac.when.empty()) {
		os<<" [";
		unsigned N = ac.when.size(),
				 dN = ac.delta.size();
		for (unsigned n=0; n<N; ++n) {
			if (n>=printMax) {
				os<<"... ("<<N-printMax<<" more)";
				break;
			}
			os<<ac.when[n].first<<'/'<<ac.when[n].second;
			if (n<dN)
				os<<'('<<setprecision(4)<<ac.delta[n]<<')';
			if (n<N-1)
				os<<", ";
		}
	}
	if (!ac.when.empty())
		os<<"]";

	return os;
} // operator<<
