//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeCorrelator.cc	--- implementation of TimeCorrelator class
//
//	DESCRIPTION:
//		Implementation part of TimeCorrelator class
//
//	CHANGE HISTORY:
//		UL, 1998-07-16: 1.0.0: - first version
//		UL, 1999-03-04: 1.1.0: - necessary changes related to lowmem Dal
//		UL, 1999-04-28: 1.2.0: - changed fatal -> error
//		UL, 1999-08-20: 1.3.0: - added more diagnostic messages
//		UL, 2000-01-11: 1.4.0: - brought in line with ODF ICD issue 2.4
//								 (FTFINE column of TCS is now unsigned)
//		UL, 2001-03-02: 1.5.0: - implemented recalculation of signal
//								 propagation delays (using column ERTIMRAW
//								 in TCS); option to switch to either scheme
//								 via environment variable SAS_OBT_MET_FIT
//							   - bug fix: TCS file was not closed
//		UL, 2001-03-30: 1.3.0: - added
//									TimeCorrelator::toOnBoardTime()
//							   - give some routines better names
//		UL, 2001-04-04: 1.4.0: - construct MET-OBT fit polynomial
//								 with original OBT-MET poly specs
//		UL, 2001-05-31: 1.4.1: - ignore entries with UTCDAY==0 in TCS
//		UL, 2001-07-24: 1.5.0: - ignore also entries with UTCFLAG!=0 and
//								 all entries after an apparent OBT reset
//								 (see comments below); fixes problem of
//								 SPR 2127
//		UL, 2001-12-10: 1.6.0: - merged two TimeCorrelator construction
//								 methods into one
//		UL, 2002-01-03: 1.7.0: - make robust against non-numeric GSID values
//								 from TCS
//							   - added TimeCorrelator::constructionMethod()
//		UL, 2002-03-25: 1.8.0: - if number of TC data points is now less than
//								 5 times the requested degree of polynomial
//								 revert back to linear fit
//							   - warning if OBT-MET polynomial is evaluated
//								 outsise abscissa data validity range
//
//	AUTHOR:
//		UL, Thu Jul 16 10:40:31 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: ICD: Observation and Slew Data Files, XMM-SOC-ICD-0004-SSD,
//			 Issue 1.0, K. Galloway
//
//	$Header: /usr/local/cvs/sasdev/sas/packages/oal/src/TimeCorrelator.cc,v 1.37 2002/03/25 16:07:49 ulammers Exp $
//

#include "TimeCorrelator.h"
#include "CalOalUtility.h"
#include "OalServer.h"
#include "OalDalUtils.h"
#include "PhysicalConstants.h"
#include "Numeric.h"
#include "Dal.h"
#include "Tracer.h"
#include "slatec.h"
#include "errstr.h"

#include <iomanip>
#include <algorithm>

using namespace CalOalUtility;
using namespace std;

//
//	info class which encapsulates access to environment variable
//	SAS_OBT_MET_LINEAR; if defined it enforces a linear fit the to
//	time correlation (TC) data using a given slope (first field) and
//	a single data point (nrows*fraction (second field) from the
//	TCS
//
class LinearObtMet {
public:
	LinearObtMet() : _enabled(false), _slope(1.), _fraction(0.) {
		char *specPtr;
		if ((specPtr = getenv("SAS_OBT_MET_LINEAR"))) {
			_enabled = true;
			vector<string> v(breakup(specPtr, ","));
			if (v.size() < 1 || v.size() > 2)
				errstr<<"Error in decoding `"<<specPtr<<"' - syntax is: "
						"slope,[fraction]"
					  <<Sas::error("FormatError");

			_slope = toNumber<double>(v[0]);
			_fraction = toNumber<double>(v[1]);
			if (_fraction<0. || _fraction>1.)
				errstr<<"Fraction fields in `"<<specPtr<<"' must lie in "
						"interval [0, 1]"
					  <<Sas::error("ValueOutOfValidBounds");
		}
	}

	bool enabled() const { return _enabled; }
	double slope() const { return _slope; }
	double fraction() const { return _fraction; }

	~LinearObtMet() {}

private:
	bool	_enabled;
	double	_slope, _fraction;
};

const FitPolynomial *TimeCorrelator::doPolyFit(FitPolySpecs &fitPolySpecs,
											   vector<double> &x,
											   vector<double> &y) const
{
	#define MESSAGE Sas::message(Msg::Lib, Msg::Verbose);

	unsigned nPoints = x.size(), pointsRemoved = 0;

	//	require that number of abscissa points is at least 5 times the
	//	requested max. degree of fit polynomial - if not revert to
	//	linear fit and issue warning
	if (fitPolySpecs.degree > 1 && nPoints < 5*fitPolySpecs.degree) {
		errstr<<"Number of time correlation data points ("<<nPoints
			  <<") is less than 5 times the requested maximum degree of the "
				"OBT-MET fit polynomial ("<<fitPolySpecs.degree
			  <<") - reverting to linear fit with RMS=1.e-2"
			  <<Sas::warning("TooFewTimeCorrelationDataPoints");
		fitPolySpecs.degree = 1;
		fitPolySpecs.rms = 1.e-2;
	}

	FitPolynomial *fitPoly = new FitPolynomial(x, y, fitPolySpecs.degree,
											   fitPolySpecs.rms);
	double removedFrac = 0.;
	while (x.size() > fitPolySpecs.degree &&
		   removedFrac <= fitPolySpecs.discardPoints &&
		   fitPolySpecs.rms < fitPoly->rms()) {
		//
		//	no satisfactory fit found yet - delete fit polynomial,
		//	remove "worst" data point and re-calculate
		//
		vector<double> diff(x.size());
		for (unsigned n=0; n<x.size(); ++n)
			diff[n] = fabs(y[n] - fitPoly->evaluate(x[n]));
		vector<double>::iterator pos = max_element(diff.begin(), diff.end());
		x.erase(pos-diff.begin()+x.begin());
		y.erase(pos-diff.begin()+y.begin());
		removedFrac = DBL(++pointsRemoved)/DBL(nPoints);

		delete fitPoly;
		fitPoly = new FitPolynomial(x, y,fitPolySpecs.degree,fitPolySpecs.rms);
	}

	if (removedFrac > fitPolySpecs.discardPoints)
		errstr<<"No satisfactory fit found after removing "
			  <<setprecision(2)<<100.*removedFrac<<"% of the data points"
			  <<Sas::error("UnsuccessfulFit");

	if (x.size() <= fitPolySpecs.degree || fitPolySpecs.rms < fitPoly->rms())
		errstr<<"Could not find a satisfactory fit"
			  <<Sas::error("UnsuccessfulFit");

	//
	//	be informative
	//
	vector<double> diff(x.size());
	for (unsigned n=0; n<x.size(); ++n)
		diff[n] = y[n] - fitPoly->evaluate(x[n]);

	errstr<<"Fit polynomial found - degree: "<<fitPoly->degree()
		  <<" RMS: "<<setprecision(5)<<fitPoly->rms()
		  <<MESSAGE;
	errstr<<"Differences between data points and fit polynomial:";
	for (unsigned n=0; n<diff.size(); ++n)
		errstr<<"   "<<setprecision(16)<<x[n]<<" "<<setprecision(16)<<diff[n]
			  <<MESSAGE;

	fitPolySpecs.degree = fitPoly->degree();
	fitPolySpecs.rms = fitPoly->rms();
	fitPolySpecs.discardPoints = removedFrac;

	return fitPoly;
	
	#undef MESSAGE
} // TimeCorrelator::doPolyFit()


void TimeCorrelator::construct(const string &timeCorrelationFileName,
							   bool recalcDelays)
{
	errstr<<"Opening Time Correlation file `"<<timeCorrelationFileName
		  <<"' ..."<<Sas::message(Msg::Lib, Msg::Verbose);
	DataSet *set = dataSetServer->open(timeCorrelationFileName, DataSet::Read);
	Table *table = set->table(0);
	unsigned nRows = table->rows();

	//
	//	check if we have column ERTIMRAW - if not this is an old-style
	//	TCS produced before Dec 2000
	//
	if (recalcDelays && !table->hasColumn("ERTIMRAW")) {
		errstr<<"Table `"<<table->qualifiedName()<<"' has no column ERTIMRAW "
				"which is needed for the recalculation of time correlation "
				"data - will continue with XMCS generated values"
			  <<Sas::warning("NoSuchColumn");
		recalcDelays = false;
	}

	ColumnData *erttimecol = table->column("ERTIME")->data(),
			   *cdobtcoarse	= table->column("OBTCOARS")->data(),
			   *cdobtfine = table->column("OBTFINE")->data(),
			   *cdutcday = table->column("UTCDAY")->data(),
			   *cdutcmilli = table->column("UTCMILLI")->data(),
			   *cdutcmicro = table->column("UTCMICRO")->data(),
			   *cdutcflag = table->column("UTCFLAG")->data(),
			   *cdgsid = table->column("GSID")->data(),
			   *cdertraw = recalcDelays ? table->column("ERTIMRAW")->data(): 0;

	int32	*obtcoarse	= cdobtcoarse->int32Data();
	uint16	*obtfine	= cdobtfine->uint16Data();
	int16	*utcday		= cdutcday->int16Data();
	int32	*utcmilli	= cdutcmilli->int32Data();
	int16	*utcmicro	= cdutcmicro->int16Data(),
			*utcflag	= cdutcflag->int16Data();

	vector<double>	obt, met, ert;

	//
	//	get very first UTC field and establish _utcBaseTime
	//
	_utcBaseTime = STime(1958, 1, 1+utcday[0]);
	_utcOffset = _utcBaseTime - _baseTime;

	errstr<<"TimeCorrelator::utcBaseTime = "<<_utcBaseTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"TimeCorrelator::baseTime    = "<<_baseTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);
	errstr<<"TimeCorrelator::utcOffset   = "<<setprecision(12)
		  <<_utcOffset
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	unsigned lastObtCoarse = OalDalUtils::unsigned32(obtcoarse[0]),
			 nWraps = 0, nskips = 0;

	//
	//	check that we have at least two rows and get inital GS ID
	//
	if (!nRows)
		errstr<<"The Time Correlation data set `"<<set->name()
			  <<"' in the current ODF is empty"
			  <<Sas::fatal("NoTimeCorrelationData");

	const string gsid = stringColumnRow(cdgsid, 0);
	for (unsigned n=0; n<nRows; ++n) {
		if (utcday[n] <= 0 || utcflag[n]) {
			nskips++;
				continue;	//	ignore entries with
							//		* UTCDAY==0
							//		* UTCFLAG!=0
							//	note: some early ODFs e.g. 0078/0124100101 have
							//	UTCDAY==0 but UTCFLAG==0 so, UTCFLAG is not
							//	a reliable quality indicator
		}
		if (stringColumnRow(cdgsid, n) != gsid)
			//
			//	change in GS - see if we want to ignore the remaining data
			//
			if (!_obtMetFitPolySpecs.ignoreGSHandovers)
				break;

		//
		//	get OBT value
		//
		unsigned uobtcoarse = OalDalUtils::unsigned32(obtcoarse[n]);
		double diff = INT(uobtcoarse) - INT(lastObtCoarse);
		if (diff < 0.)
			if (-diff > .9*65536.*65536.)
				//
				//	coarse time must have had a wrap around
				//
				++nWraps;
			else {
				//
				//	this is an OBT reset (e.g. ODF 102_0099020301 jump
				//	from 1114402 to 33 at TCS record 2827) - cannot use
				//	subsequent time	correlation entries; ignore the rest
				//
				errstr<<"Detected apparent reset of onboard clock in time "
						"correlation data table `"<<table->qualifiedName()
					  <<" between records #"<<n<<'/'<<n+1
					  <<" where OBTCOARSE jumps from "
					  <<lastObtCoarse<<" to "<<uobtcoarse<<" - ignoring all "
						"subsequent records. The impact of this on "
						"the accuracy of event arrival times is very likely "
						"to be negligible. If you intend to perform "
						"high-accuracy timing analysis on this ODF data set, "
						"please consult the XMM-Newton helpdesk and ask for "
						"confirmation of this assessment. In doing so please "
						"detail the complete ODF designation and the version "
						"of the SAS you are using at this moment."
					  <<Sas::warning("OBTreset");
				break;
			}

		lastObtCoarse = uobtcoarse;
		obt.push_back(DBL(uobtcoarse) +
						 DBL(OalDalUtils::unsigned32(obtfine[n]))/65536. +
						 DBL(nWraps) * 65536. * 65536.);

		//
		//	get UTC value
		//
		if (!recalcDelays)
			met.push_back(STime(1958, 1, 1+utcday[n], 0, 0, 0,
						 1.e-3 * DBL(utcmilli[n]) + 1.e-6 * DBL(utcmicro[n])) -
						_utcBaseTime);
		else
			met.push_back(metWithPropDelays(stringColumnRow(cdertraw, n),
											gsid));

		//
		//	get ERT value
		//
		ert.push_back(STime(stringColumnRow(erttimecol, n)) - _utcBaseTime);

//		errstr<<setprecision(12)
//			  <<n<<" "<<obt[n-nskips]-obt[0]<<" "<<met[n-nskips]-met[0]
//			  <<" "<<ert[n-nskips]-ert[0]
//			  <<Sas::message(Msg::Lib, Msg::Verbose);
		errstr<<setprecision(12)
			  <<obt[n-nskips]-obt[0]<<" "<<met[n-nskips]-met[0]
			  <<Sas::message(Msg::Lib, Msg::Verbose);
	}

	//	tidy up and close TCS
	delete erttimecol;
	delete cdertraw;
	delete cdobtcoarse;
	delete cdobtfine;
	delete cdutcday;
	delete cdutcmilli;
	delete cdutcmicro;
	delete cdutcflag;
	delete cdgsid;

	dataSetServer->close(set);

	//
	//	Do we want to enforce a linear OBT-MET relation?
	//
	LinearObtMet lininfo;
	if (lininfo.enabled() || obt.size() == 1) {
		double slope = 1.;
		if (lininfo.enabled()) {
			slope = lininfo.slope();
			unsigned nPt = UNT(obt.size()*lininfo.fraction());
			double obt0 = obt[nPt],
				   met0 = met[nPt],
				   ert0 = met[nPt];
			obt.clear(); obt.push_back(obt0);
			met.clear(); met.push_back(met0);
			ert.clear(); ert.push_back(ert0);
		} else
			errstr<<"Only one usable time correlation data point in TCS "
					"of current ODF - will assume perfect linear relationship "
					"between OBT and MET"
				  <<Sas::warning("AssumedOBT_METrelation");

		obt.push_back(obt.back()+1.);
		ert.push_back(ert.back()+1.);
		met.push_back(met.back()+slope);
	}

	//
	//	calculate fit polynomials
	//
	vector<double> met2(met);
	FitPolySpecs obtMetFitPolySpecs = _obtMetFitPolySpecs;
	_obtMetPoly = doPolyFit(_obtMetFitPolySpecs, obt, met);
	_metObtPoly = doPolyFit(obtMetFitPolySpecs, met, obt);
	_ertMetPoly = doPolyFit(_ertMetFitPolySpecs, ert, met2);

	errstr<<"OBT-MET fit polynomial successfully calculated - signature:\n"
		  <<*_obtMetPoly
		  <<Sas::message(Msg::Lib, Msg::Sparse);
} // TimeCorrelator::construct()


double TimeCorrelator::metWithPropDelays(const string &ertraw,
										 const string &gsid)
{
	static const PositionLocator *posloc = 0;

	if (!posloc)
		posloc = oalServer.positionLocator();

	unsigned short gsi = 0;
	try {
		gsi = toNumber<unsigned short>(gsid);
	} catch(...) {
		errstr<<"Cannot convert GSID `"<<gsid
			  <<"' to valid numeric ground station identifier"
			  <<Sas::error("InvalidGSID");
	}

	const XmmGroundStation &gs = XmmMission::groundStation(gsi);
	STime t(ertraw);
	CartesianVector xmmPos = posloc->stateVector(t).position,
					gsPos = gs.stateVector(t).position;
	double met,
		   xmmGSdist = (gsPos - xmmPos).norm() * 1000.,	//	km -> m
		   propDelay =
				//	OBDH delay
				XmmMission::dataHandler().timingDelay() +

				//	signal travel time - NOTE: atmospheric propagation
				//	delays are of the order of 0.1 us (see email
				//	F. Dreger dated 2001-02-28), so, negligible
			    xmmGSdist/PhysicalConstants::c +

				// Ground station delays
				gs.timingDelay();

	errstr<<"Total propagation delay at T="<<t.toString(6)
		  <<": "<<propDelay<<"s"
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	met = t - _utcBaseTime;
	met -= propDelay;

	return met;
} // TimeCorrelator::metWithPropDelays()


TimeCorrelator::TimeCorrelator(const STime &timeBase)
	: OalChild<OalServer>(0), _baseTime(timeBase), _lastObtTag(0.),
							  _obtWraps(0), _obtMetPoly(0), _ertMetPoly(0)
{
	_utcBaseTime = STime::earliest;

	_obtMetFitPolySpecs.degree = 0;
	_obtMetFitPolySpecs.rms = 0.;
	_obtMetFitPolySpecs.discardPoints = 0;

	_ertMetFitPolySpecs.degree = 0;
	_ertMetFitPolySpecs.rms = 0.;
	_ertMetFitPolySpecs.discardPoints = 0;
} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(const OdfSet *timeCorrelationSet,
						   const FitPolySpecs &obtMetFitPolySpecs,
						   const FitPolySpecs &ertMetFitPolySpecs,
						   ConstructionMethod cm,
						   const STime &timeBase)
	: OalChild<OalServer>(0), _obtMetFitPolySpecs(obtMetFitPolySpecs),
	  _ertMetFitPolySpecs(ertMetFitPolySpecs),
	  _constructionMethod(cm),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{
	construct(timeCorrelationSet->name(), cm == TCS_RECALCULATE_DELAYS);
} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(const string &timeCorrelationSetName,
						   const FitPolySpecs &obtMetFitPolySpecs,
						   const FitPolySpecs &ertMetFitPolySpecs,
						   ConstructionMethod cm,
						   const STime &timeBase)
	: OalChild<OalServer>(0), _obtMetFitPolySpecs(obtMetFitPolySpecs),
	  _ertMetFitPolySpecs(ertMetFitPolySpecs),
	  _constructionMethod(cm),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{
	construct(timeCorrelationSetName, cm == TCS_RECALCULATE_DELAYS);
} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::TimeCorrelator(const OalServer *oalEngine,
					const FitPolySpecs &obtMetFitPolySpecs,
					const FitPolySpecs &ertMetFitPolySpecs,
					ConstructionMethod cm,
					const STime &timeBase)
	: OalChild<OalServer>(oalEngine), _obtMetFitPolySpecs(obtMetFitPolySpecs),
	  _ertMetFitPolySpecs(ertMetFitPolySpecs),
	  _constructionMethod(cm),
	  _baseTime(timeBase), _lastObtTag(0.), _obtWraps(0)
{
	assert(oalEngine);
	construct(oalEngine->odf()->selectSet(Xmm::SC, Xmm::TIME_CORRELATION, 1)
				->name(), cm == TCS_RECALCULATE_DELAYS);
} // TimeCorrelator::TimeCorrelator()


TimeCorrelator::ConstructionMethod TimeCorrelator::obtMetFitPolyMethod() const
{
	return _constructionMethod;
} // TimeCorrelator::obtMetFitPolyMethod()


const TimeCorrelator::FitPolySpecs &TimeCorrelator::obtMetFitPolySpecs() const
{
	return _obtMetFitPolySpecs;
} // TimeCorrelator::obtMetFitPolySpecs()


const TimeCorrelator::FitPolySpecs &TimeCorrelator::ertMetFitPolySpecs() const
{
	return _ertMetFitPolySpecs;
} // TimeCorrelator::obtMetFitPolySpecs()


void TimeCorrelator::setBaseTime(const STime &) const
{
	errstr<<"TimeCorrelator::setBaseTime is not yet implemented"
		  <<Sas::error("NotYetImplemented");
} // TimeCorrelator::setBaseTime()


double TimeCorrelator::toMissionEllapsedTime(double obtTag) const
{
	if (obtTag < _lastObtTag &&			//	a time decrease
		_lastObtTag - obtTag > 320000.)	//	and difference is large enough
										//	to believe a wrap-around has
										//	occured
		++_obtWraps;

	_lastObtTag = obtTag;
	obtTag += DBL(_obtWraps) * 65536. * 65536.;

	return _obtMetPoly->evaluate(obtTag) + _utcOffset;
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toMissionEllapsedTime(unsigned obtCoarse,
											 unsigned short obtFine) const
{
	return toMissionEllapsedTime(DBL(obtCoarse) + DBL(obtFine)/65536.);
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toOnBoardTime(double missionEllapsedTime) const
{
	return _metObtPoly->evaluate(missionEllapsedTime - _utcOffset);
} // TimeCorrelator::toOnBoardTime()


void TimeCorrelator::resetObtMetRelation() const
{
	_lastObtTag = 0.;
	_obtWraps = 0;
} // TimeCorrelator::resetObtMetRelation()


void TimeCorrelator::plotObtMetRelation() const
{
	errstr<<"TimeCorrelator::plotObtMetRelation is not yet implemented"
		  <<Sas::error("NotYetImplemented");
} // TimeCorrelator::plotObtMetRelation()


double TimeCorrelator::toMissionEllapsedTime(const string &ertString) const
{
	return toMissionEllapsedTime(STime(ertString));
} // TimeCorrelator::toMissionEllapsedTime()


double TimeCorrelator::toMissionEllapsedTime(const STime &ertTag) const
{
	double arg = ertTag - _utcBaseTime;

	return _ertMetPoly->evaluate(arg) + _utcOffset;
} // TimeCorrelator::toMissionEllapsedTime()


void TimeCorrelator::plotErtMetRelation() const
{
	errstr<<"TimeCorrelator::plotErtMetRelation is not yet implemented"
		  <<Sas::error("Not yet implemented");
} // TimeCorrelator::plotErtMetRelation()

namespace {
	template<class T>
	void addAttrib(Attributable *attrib,
				   const string &name, const T &val, const string &comment) {
		Attribute *att = attrib->hasAttribute(name)
							? attrib->attribute(name)
							: attrib->addAttribute(name, comment);
		*att = val;
	}
}

const TimeCorrelator &TimeCorrelator::addCommonAttributes(
												Attributable *attrib) const
{
	addAttrib(attrib, "TFIT_RPD",
			  obtMetFitPolyMethod()==TCS_RECALCULATE_DELAYS,
			  "Recalculated signal propagation delays?");
	const FitPolySpecs &specs = obtMetFitPolySpecs();
	addAttrib(attrib, "TFIT_DEG", int(specs.degree),
			  "Degree of OBT-MET fit polynomial");
	addAttrib(attrib, "TFIT_RMS", specs.rms,
			  "RMS value of OBT-MET polynomial fit");
	addAttrib(attrib, "TFIT_PFR", specs.discardPoints,
			  "Fraction of disregarded TCS data points");
	addAttrib(attrib, "TFIT_IGH", specs.ignoreGSHandovers,
			  "Ignored GS handover in TC data?");

	return *this;
} // TimeCorrelator::addCommonAttributes()


TimeCorrelator::~TimeCorrelator()
{
	delete _obtMetPoly;
	delete _ertMetPoly;
} // TimeCorrelator::~TimeCorrelator()



//
//	---------------------( FitPolynomial )-----------------------------------
//
FitPolynomial::FitPolynomial(const vector<double> &x,
							 const vector<double> &y,
							 const unsigned maxDegree, double rms)
	: _x(x), _y(y)
{
	if (rms <= 0. || maxDegree == 0)
		errstr<<"RMS must be >= 0 and maximum allowed degree of fit polynomial"
				" must be >0"<<Sas::error("IllegalValues");

	unsigned nPoints = x.size();
	if (nPoints > maxDegree) {
		double	*weights = new double [nPoints],
				*vals = new double [nPoints],
				*xp = new double [nPoints],
				*yp = new double [nPoints];
		for (unsigned n=0; n<nPoints; ++n) {
			xp[n] = x[n]; yp[n] = y[n];
		}
		_fitPoly	= new double [3*nPoints+3*maxDegree+3];
		weights[0]	= -1.;		// do an unweighted least-square-fit
		_rms		= 1.e-20;	// polynomal fit will not achieve this RMS,
								// thus, we will always use the maximum
								// possible degree
		int ierr, degree;		//	error code from SLATEC

        SLATEC_DPOLFT(nPoints, xp, yp, weights, maxDegree, degree, _rms,
					  vals, ierr, _fitPoly);
		_degree = degree;

		if (ierr != 1 && ierr != 3)
			//
			//	fit failed
			//
			errstr<<"Routine `DPOLFIT' failed to find a polynomal fit of "
					"maximum degree "<<maxDegree
				  <<" - return code is "<<ierr
				  <<Sas::error("UnsuccessfulFit");

		//
		//	fit is ok
		//
		errstr<<"Fit polynomial computed - optimized degree: "<<_degree
			  <<" - RMS error: "<<_rms
			  <<Sas::message(Msg::Lib, Msg::Verbose);
		delete [] weights;
		delete [] vals;
		delete [] xp;
		delete [] yp;
	} else
		errstr<<"Only "<<nPoints<<" data points are available - cannot do a "
				"fit to a polynomial of degree "<<maxDegree
			  <<Sas::error("TooFewDataPoints");
} // FitPolynomial::FitPolynomial()


double FitPolynomial::evaluate(double x, bool warningOnExtrapolation) const
{
	double retVal;

	if (warningOnExtrapolation && (x<_x.front()-100. || x>_x.back()+100.))
		//	issue extrapolation warning
		errstr<<"Evaluating OBT-MET fit polynomial at x="<<setprecision(10)
			  <<x<<" which is outside the time correlation data abscissa "
				"range ["<<setprecision(10)<<_x.front()<<", "<<_x.back()
			  <<"] - result of extrapolation might be afflicted with large "
				"numerical uncertainty."<<Sas::warning("Extrapolation");

	SLATEC_DP1VLU(_degree, 0, x, retVal, static_cast<double *>(0), _fitPoly);

	return retVal;
} // FitPolynomial::evaluate()


double FitPolynomial::derivative(double x, unsigned short n,
								 bool warningOnExtrapolation) const
{
	if (n==0)
		return evaluate(x, warningOnExtrapolation);

	double retVal;
	double deriv[n];
	int di = INT(n);

	SLATEC_DP1VLU(_degree, di, x, retVal, deriv, _fitPoly);

	return deriv[n-1];
} // FitPolynomial::derivative()


FitPolynomial::~FitPolynomial()
{
	delete [] _fitPoly;
} // FitPolynomial::~FitPolynomial()


ostream &operator<<(ostream &os, const FitPolynomial &fp)
{
	unsigned N=fp.degree();
	os<<"# data points: "<<fp.xPoints().size()<<endl
	  <<"degree       : "<<N<<endl
	  <<"RMS          : "<<setprecision(8)<<fp.rms()<<endl
	  <<"coefficients : "<<endl;
	for (unsigned n=0; n<=N; ++n)
		os<<"    a"<<n<<" : "
		  <<setprecision(12)<<fp.derivative(0., n, false)/
												Numeric::factorial(n)<<endl;
	return os;
} // operator<<
