//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	XmmPositionLocator.cc	--- implementation of XmmPositionLocator classes
//
//	DESCRIPTION:
//		Implementation part of XmmPositionLocator classes
//
//	CHANGE HISTORY:
//		UL, 1998-08-12: 1.0.0: - first version
//		UL, 1999-01-27: 1.1.0: - minor changes to avoid gcc 2.8.1 warnings
//		UL, 1999-04-28: 1.2.0: - fatal -> error
//		UL, 1999-06-11: 1.3.0: - added 
//		UL, 1999-08-18: 1.6.0: - added exception handling
//		UL, 2000-02-24: 1.7.0: - added proper implementation based on
//								 'orbita' routine from FD
//		UL, 2000-03-06: 1.8.0: - reading of orbit file is more robust w.r.t to
//								 recoginition of EOF conditions
//		UL, 2000-11-02: 1.9.0: - various modifications
//
//	AUTHOR:
//		U. Lammers, Wed Aug 12 14:02:33 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: ICD: Observation and Slew Data Files, XMM-SOC-ICD-0004-SSD,
//			 Issue 1.0, K. Galloway
//		[3]: ICD: XMCS Orbit Access Software, XMM-MOC-ICD-0021-OAD, issue 1.1
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/XmmPositionLocator.cc,v 1.7 2002/02/06 13:05:48 ulammers Exp $
//

#include <fstream>
#include <cassert>
#include <cmath>
#include <iomanip>

#include "XmmPositionLocator.h"
#include "Interpolator.h"
#include "CalOalUtility.h"
#include "Celestial.h"
#include "OalUtils.h"
#include "OalServer.h"
#include "Xmm.h"
#include "Dal.h"
#include "Tracer.h"
#include "errstr.h"
#include "common.h"

using namespace CalOalUtility;

class OrbitSet {
public:
	OrbitSet(const string &name) : _name(name)
	{
		_f = new ifstream(name.c_str());
		if (!*_f)
			errstr<<"Cannot open orbit file `"<<name
				  <<Sas::error("OrbitFileOpenError");
	}	

	bool nextEntry(XmmPositionLocator::OrbitData &od) const
	{
		skipROL();	//	skip first record
		checkEof();
		unsigned val;

		//	second record
		*_f>>val;	//	record ID
		double valBeg, epoch, valEnd;
		*_f>>valBeg;				//	start of validity
		*_f>>valEnd;				//	end of validity
		*_f>>epoch;					//	epoch of reference Kepler orbit
		od.startValidity = mjd2000(valBeg);
		od.endValidity = mjd2000(valEnd);
		od.refEpoch = mjd2000(epoch);

		*_f>>od.revolution;		//	revolution number
		*_f>>od.a;				//	semimajor axis
		*_f>>od.invMeanMotion;	//	inverse mean motion
		skipROL();
		checkEof();

		//	third record
		*_f>>val;		//	record ID
		od.nCoef = val - 300;	//	see [3]		
		for (unsigned n=0; n<NUM_OF(od.stateVector);	//	state vector
				*_f>>od.stateVector[n++]);			//	of reference orbit
		*_f>>od.dist;	//	distance S/C-earth center	
		skipROL();

		for (unsigned i=0; i<od.nCoef; ++i) {
			*_f>>val;				//	record ID
			for (unsigned j=0; j<6; ++j)
				*_f>>od.chebCoef[i][j];
			skipROL();
		}

		return !_f->eof();
	}

	~OrbitSet() { delete _f; _f = 0; }

private:
	ifstream *_f;
	string	 _name;

	//
	//	skip rest-of-line
	//
	inline void skipROL() const { _f->ignore(100, '\n'); }

	//
	//	check for unexpected EOF
	//
	inline void checkEof() const {
		if (_f->eof())
			errstr<<"Unexpected EOF in "<<_name<<Sas::error("UnexpectedEOF");
	}

	//
	//	compute number of seconds since mission reference time from
	//	MJD2000 date
	//
	double mjd2000(double t) const
	{
		static double mjdY2K = STime(2000, 1, 1).modifiedJulianDay().full();

		return STime(t + mjdY2K) - Xmm::missionReferenceTime;
	}
}; // OrbitSet


//
//	---( XmmPositionLocator )------------------------------------------------
//
void XmmPositionLocator::construct(const string &orbitSetName,
								   const STime &startTime,
								   const STime &endTime)
{
	errstr<<"Opening orbit set `"<<orbitSetName<<"' and reading data "
			"covering the period "<<startTime<<" - "<<endTime
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	OrbitSet o(orbitSetName);
	double tstart = startTime - Xmm::missionReferenceTime,
		   tend = endTime - Xmm::missionReferenceTime;

	//	find entry with validity start date after start time
	OrbitData od;
	bool startFound = false;
	while (o.nextEntry(od))
		if (od.endValidity >= tstart) {
			startFound = true;
			break;
		}

	if (!startFound)
		errstr<<"Could not find parameters in orbit file `"
			  <<orbitSetName<<"' which are valid at given start time "
			  <<startTime<<" ("<<tstart<<")"
			  <<Sas::error("NoOrbitData");

	_orbitData.push_back(od);
	bool neof;
	do {
		if (neof = o.nextEntry(od))
			_orbitData.push_back(od);
	} while (neof && od.endValidity < tend);

	unsigned N = _orbitData.size();
	if (_orbitData[N-1].endValidity < tend)
		errstr<<"Orbit file ended before given end time "
			  <<endTime<<" ("<<tend<<")"
			  <<Sas::warning("PrematureEndOfOrbitFile");
} // XmmPositionLocator::construct()


XmmPositionLocator::XmmPositionLocator(const OalServer *oalEngine,
									   unsigned cm,
									   double maxDelta)
	: PositionLocator(Xmm::missionReferenceTime),
	  OalChild<OalServer>(oalEngine), _maxDelta(maxDelta),
	  _positionDataSource(0)
{
	const OdfSet *orbitSet = 0;

	bool hasReconOrbit =
			parent()->odf()->exposures(Xmm::SC, Xmm::RECONSTRUCTED_ORBIT),
		 hasPredOrbit =
			parent()->odf()->exposures(Xmm::SC, Xmm::PREDICTED_ORBIT);

	if (cm & FROM_RECONSTRUCTED_ORBIT)
		if (hasReconOrbit) {
			orbitSet = parent()->odf()->selectSet(Xmm::SC,
												  Xmm::RECONSTRUCTED_ORBIT, 1);
			_positionDataSource = FROM_RECONSTRUCTED_ORBIT;
		} else {
			errstr<<"The ODF in `"<<oalEngine->odf()->directoryName()
				  <<"' does not contain a reconstructed orbit file.";
			if (cm & FROM_PREDICTED_ORBIT) {
				errstr<<" Using predicted orbit file."
					  <<Sas::warning("NoReconstructedOrbitFile");
				if (hasPredOrbit) {
					orbitSet = parent()->odf()->selectSet(Xmm::SC,
												  Xmm::PREDICTED_ORBIT, 1);
					_positionDataSource = FROM_PREDICTED_ORBIT;
				} else
					errstr<<"The ODF does neither contain a predicted orbit "
							"file - cannot continue"
						  <<Sas::error("NoOrbitData");
			} else
				errstr<<Sas::error("NoReconstructedOrbitFile");
		}
	else if (cm & FROM_PREDICTED_ORBIT && hasPredOrbit) {
		orbitSet = parent()->odf()->selectSet(Xmm::SC,
											  Xmm::PREDICTED_ORBIT, 1);
		_positionDataSource = FROM_PREDICTED_ORBIT;
	} else
		errstr<<"The ODF in `"<<oalEngine->odf()->directoryName()
			  <<"' does not contain a predicted orbit file.";

	if (!orbitSet)
		errstr<<Sas::error("NoOrbitData");

	construct(orbitSet->name(), parent()->odf()->observationStartTime(),
			  parent()->odf()->observationEndTime());
} // XmmPositionLocator::XmmPositionLocator()


XmmPositionLocator::XmmPositionLocator(const OdfSet *orbitSet,
									   const STime &startTime,
									   const STime &endTime,
									   double maxDelta)
	: PositionLocator(Xmm::missionReferenceTime),
	  OalChild<OalServer>(0), _maxDelta(maxDelta),
	  _positionDataSource(0)
{
	if (!orbitSet)
		errstr<<"No orbit file specified"<<Sas::error("NoOrbitFile");
	construct(orbitSet->name(), startTime, endTime);
} // XmmPositionLocator::XmmPositionLocator()


XmmPositionLocator::XmmPositionLocator(const string &orbitSetName,
									   const STime &startTime,
									   const STime &endTime,
									   double maxDelta)
	: PositionLocator(Xmm::missionReferenceTime),
	  OalChild<OalServer>(0), _maxDelta(maxDelta)
{
	construct(orbitSetName, startTime, endTime);
} // XmmPositionLocator::XmmPositionLocator()


XmmPositionLocator::OrbitDataIter XmmPositionLocator::find(double time,
		XmmPositionLocator::OrbitDataIter startHint) const
{
	OrbitDataIter startIter[] = { startHint, _orbitData.begin() };

	for (unsigned n=0; n<NUM_OF(startIter); ++n)
		for (OrbitDataIter i = startIter[n]; i != _orbitData.end(); ++i)
			if ((*i).startValidity <= time && (*i).endValidity >= time) {
				//	found the parameter set
				return i;
		}

	//
	//	not found - the requested time lies either in a gap for before
	//	the start of the first or after the end of the last record
	//
	OrbitDataIter i = _orbitData.begin();
	while (i != _orbitData.end() && time > (*i).startValidity)
		++i;
	double tdist;
	if (i != _orbitData.begin())
		if (i != _orbitData.end())
			if ((*i).startValidity - time > time - (*(i-1)).endValidity) {
				--i;
				tdist = time - (*i).endValidity;
			} else
				tdist = (*i).startValidity - time;
		else {
			--i;
			tdist = time - (*i).endValidity;
		}
	else
		tdist = (*i).startValidity - time;

	//
	//	check if we are at most _maxDelta away from closest set
	//
	if (tdist <= _maxDelta) {
		errstr<<"The requested time is less than "<<_maxDelta<<"s outside "
				"the validity range of the closest orbit parameter set ("
			  <<tdist<<"s) - performing extrapolation with closest set"
			  <<Sas::warning("TimeOutsideValidityRange");
		return i;
	}

	return _orbitData.end();
} // XmmPositionLocator::find()



XmmPositionLocator::~XmmPositionLocator()
{
} // XmmPositionLocator::~XmmPositionLocator()


//
//	---( XmmPositionLocator::stateVector )------------------------------------
//
StateVector XmmPositionLocator::stateVector(const STime &time) const
{
	return stateVector(time - Xmm::missionReferenceTime);
} // XmmPositionLocator::stateVector()

StateVector XmmPositionLocator::stateVector(double time) const
{
	static OrbitDataIter lastOrbitData = _orbitData.begin();

	OrbitDataIter i = find(time, lastOrbitData);
	if (i==_orbitData.end()) {
		errstr<<"State vector requested at time "<<setprecision(12)<<time
			  <<" ("<<(Xmm::missionReferenceTime + time)<<") which is more "
				"than "<<_maxDelta
			  <<"s outside of all orbital parameter validity ranges"
			  <<Sas::warning("NoOrbitData");
		throw NoPositionDataAvailable();
	}

	lastOrbitData = i;

	//
	//	compute Kepler state vector
	//
	SimpleStateVector v;
	keplerStateVector(*i, time, v);

	//
	//	apply Chebyshev polynomial correction to Kepler state vector
	//
	chebyshevCorrect(*i, time, v);

	double meanAnom = (time - (*i).refEpoch)/(*i).invMeanMotion;
	return StateVector(v, (*i).revolution + meanAnom/M_TWO_PI);
} // XmmPositionLocator::position()


void XmmPositionLocator::keplerStateVector(const OrbitData &od, double time,
										   SimpleStateVector v) const
{
	//
	//	model Kepler orbit
	//
	double meanAnom = (time - od.refEpoch)/od.invMeanMotion;
	double arin	= od.a/od.dist,
		   arm	= (od.dist - od.a)/od.a,
		   rvwam= (od.stateVector[0]*od.stateVector[3] +
				   od.stateVector[1]*od.stateVector[4] +
				   od.stateVector[2]*od.stateVector[5]) *
					od.invMeanMotion/SQR(od.a);

	//
	//	solve Kepler's equation (M=E-e*sin(E)) by Newton iterations
	//
	double trueAnom = meanAnom - rvwam,
		   tol = 1.e-7 + 1.e-10 * fabs(trueAnom),
		   b = trueAnom,
		   g0 = 0., g1 = 0., bet = 0., d = 0.;

	static const unsigned maxIter = 15;
	unsigned n;
	for (n=0; n<maxIter; ++n) {
		g0	= cos(b);
		g1	= sin(b);
		bet	= trueAnom - arm*g1 + rvwam*g0;
		d	= (bet - b)/(1. + arm*g0 + rvwam*g1);
		b += d;
		if (fabs(d) <= tol)
			break;
	}

	if (n >= maxIter) {
		errstr<<"Kepler equation could not be solved in "<<maxIter
			  <<" iterations - cannot determine S/C state vector"
			  <<Sas::warning("NoConvergence");
		throw NoPositionDataAvailable();
	}

	g0 -= d*g1;
	g1 += d*g0;
	double g2 = 1. - g0,
		   g3 = b - g1,
		   fx = 1. - g2*arin,
		   gx = (meanAnom - g3)*od.invMeanMotion;

	//	compute position vector
	for (unsigned n=0; n<3; ++n)
		v[n] = fx*od.stateVector[n] + gx*od.stateVector[n+3];

	//	compute velocity vector
	double rx = sqrt(SQR(v[0]) + SQR(v[1]) + SQR(v[2])),
		   ft = -g1*od.a*arin/(od.invMeanMotion*rx),
		   gt = 1. - g2*od.a/rx;
	for (unsigned n=3; n<6; ++n)
		v[n] = ft*od.stateVector[n-3] + gt*od.stateVector[n];
} // XmmPositionLocator::keplerStateVector()


void XmmPositionLocator::chebyshevCorrect(const OrbitData &od, double time,
										  SimpleStateVector v) const
{
	//
	//	Chebyshev polynomial correction to Kepler state vector
	//
	if (od.nCoef > 1) {
		//	correction makes only sense for polynomials with deg > 1
		double	midTime = .5*(od.startValidity + od.endValidity),
				scale = 4./(od.endValidity - od.startValidity)*86400.;

		double	s = scale*(time - midTime)/86400.,
				pa = 1.,
				p = .5*s;

		for (unsigned n=0; n<6; ++n)
			v[n] += od.chebCoef[0][n] + p*od.chebCoef[1][n];

		if (od.nCoef > 2) {
			double pb;
			for (unsigned i=2; i<od.nCoef; ++i) {
				pb = pa;
				pa = p;
				p = s*pa - pb;
				for (unsigned n=0; n<6; ++n)
					v[n] += p*od.chebCoef[i][n];
			}
		}
	}
} // XmmPositionLocator::chebyshevCorrect()


const XmmPositionLocator &XmmPositionLocator::addCommonAttributes(
													Attributable *attrib) const
{
		static const char POS_SRC[] = "ORB_RCNS";

	Attribute *att = attrib->hasAttribute(POS_SRC)
						? attrib->attribute(POS_SRC)
						: attrib->addAttribute(POS_SRC,
										   "Reconstructed orbit data used?");
	*att = _positionDataSource == FROM_RECONSTRUCTED_ORBIT;

	return *this;
} // XmmPositionLocator::addCommonAttributes()

