//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	XmmPositionLocator.h --- classes streamlined for S/C position retrieval
//
//	DESCRIPTION:
//		A class XmmPositionLocator + auxiliary classes are defined wich
//		provide access to the S/C position information on each ODF.
//
//	CHANGE HISTORY:
//		UL, 1998-08-11: 1.0.0: - first version
//		UL, 1999-08-18: 1.1.0: - added exception class
//		UL, 2000-02-24: 1.2.0: - altered data structures in private section
//								 to reflect structure of orbit file
//		UL, 2000-11-02: 1.3.0: - modified version from PositionLocator
//
//	AUTHOR:
//		U. Lammers, Tue Aug 11 15:55:34 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: ICD: Observation and Slew Data Files (ODF-ICD), K. Galloway
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/oal/src/XmmPositionLocator.h,v 1.4 2002/01/07 16:34:51 ulammers Exp $
//

#ifndef XMM_POSITION_LOCATOR_H
#define XMM_POSITION_LOCATOR_H

#include "Tracer.h"
#include "OalChild.h"
#include "Odf.h"
#include "Celestial.h"
#include "TimeAxis.h"
#include "PositionLocator.h"

class OalServer;

class XmmPositionLocator : public PositionLocator,
						   public virtual OalChild<OalServer> {
public:
	//
	//	exception class: to be thrown if no position data is available
	//	at requested time
	//
	class NoPositionDataAvailable {};

	static const unsigned FROM_PREDICTED_ORBIT = 1,
						  FROM_RECONSTRUCTED_ORBIT = 2;
	XmmPositionLocator(const OalServer *oalEngine,
					   unsigned cm = FROM_PREDICTED_ORBIT |
									 FROM_RECONSTRUCTED_ORBIT,
					   double maxDelta = 10.);

	XmmPositionLocator(const OdfSet *orbitSet,
					   const STime &startTime,
					   const STime &endTime,
					   double maxDelta = 10.);

	XmmPositionLocator(const string &orbitSetName,
					   const STime &startTime,
					   const STime &endTime,
					   double maxDelta = 10.);

	StateVector stateVector(double time) const;
	StateVector stateVector(const STime &time) const;

	const XmmPositionLocator &addCommonAttributes(Attributable *attrib) const;
		//
		//	PURPOSE:
		//		Add boolean attribute "ORB_RCNS" to passed attributable; is
		//		true if reconstructed (otherwise predicted) orbit data
		//		are in use
		//

	~XmmPositionLocator();

private:
	//	data
	typedef double	SimpleStateVector[6];
	struct OrbitData {
		double	startValidity,	//	start of validity [s]
				refEpoch,		//	epoch of reference Kepler orbit [s]
				endValidity;	//	end of validity [s]

		double	revolution,		//	revolution number at epoch
				a,				//	semimajor axis of ref. Kepler orbit [km]
				invMeanMotion;	//	inverse mean motion of ref. orbit [s/rad]

		SimpleStateVector
				stateVector;	//	state vector at epoch [km + km/s]
		double	dist;			//	distance S/C-earth center at epoch [km]

		double	chebCoef[10][6];//	coefficients of Chebyshev polynomials
								//	for Keper orbit correction
		unsigned nCoef;			//	number of effective coefficients
	};

	double			_maxDelta;	//	maximum allowed temoral gap between
								//	OrbitData records
	unsigned		_positionDataSource;
								//	1: PREDICTED 2: RECONSTRUCTED

	typedef vector<OrbitData>	OrbitDataVector;
	OrbitDataVector _orbitData;	//	vector with orbit data
	typedef OrbitDataVector::const_iterator	OrbitDataIter;

	//	methods
	void construct(const string &, const STime &, const STime &);
	OrbitDataIter find(double time, OrbitDataIter startHint) const;
	void keplerStateVector(const OrbitData &od, double time,
						   SimpleStateVector v) const;
	void chebyshevCorrect(const OrbitData &od, double time,
						  SimpleStateVector v) const;
};

#endif
